from __future__ import annotations

import abc
import hashlib
import json
import sys
from dataclasses import dataclass
from enum import Enum, auto
from typing import (
    Any,
    Generic,
    List,
    Literal,
    Optional,
    Sequence,
    Tuple,
    TypeVar,
    Union,
    overload,
)

if sys.version_info < (3, 11):
    from typing_extensions import Self
else:
    from typing import Self

import numpy as np
import numpy.typing as npt
import sympy as sp

from .utils import CGc

__all__ = [
    "ElectronicState",
    "BasisState",
    "CoupledBasisState",
    "UncoupledBasisState",
    "State",
    "BasisStates_from_State",
    "Basis",
    "CoupledState",
    "UncoupledState",
]


class ElectronicState(Enum):
    X = auto()
    B = auto()


class Basis(Enum):
    Uncoupled = auto()
    Coupled = auto()
    CoupledP = auto()
    CoupledΩ = auto()


class Parity(Enum):
    Pos = +1
    Neg = -1


@dataclass
class BasisState(abc.ABC):
    J: int
    electronic_state: Optional[ElectronicState]
    isCoupled: bool
    isUncoupled: bool
    basis: Optional[Basis]

    # scalar product (psi * a)
    def __mul__(self, a: Union[float, complex, int]):
        raise NotImplementedError

    # scalar product (a * psi)
    def __rmul__(self, a: Union[float, complex, int]):
        raise NotImplementedError

    def __add__(self, other):
        raise NotImplementedError

    def print_quantum_numbers(self, printing: bool = False) -> str:
        raise NotImplementedError

    def transform_to_omega_basis(self):
        raise NotImplementedError

    def transform_to_parity_basis(self):
        raise NotImplementedError


class CoupledBasisState(BasisState):
    # constructor
    def __init__(
        self,
        F: int,
        mF: int,
        F1: float,
        J: int,
        I1: float,
        I2: float,
        Omega: int = 0,
        P: Optional[int] = None,
        electronic_state: Optional[ElectronicState] = None,
        energy: Optional[float] = None,
        Ω: Optional[int] = None,
        v: Optional[int] = None,
        basis: Optional[Basis] = None,
    ):
        self.F, self.mF = F, mF
        self.F1 = F1
        self.J = J
        self.I1 = I1
        self.I2 = I2

        if Ω is not None:
            self.Ω: int = Ω
            self.Omega = self.Ω
        elif Omega is not None:
            self.Omega = Omega
            self.Ω = self.Omega
        else:
            raise AssertionError("need to supply either Omega or Ω")
        if P is not None:
            self.P: Optional[int] = P
        else:
            self.P = None
        #     raise AssertionError("need to supply parity P")
        self.electronic_state = electronic_state
        self.energy = energy
        self.isCoupled = True
        self.isUncoupled = False

        # determine which basis we are in
        if basis is not None:
            self.basis = basis
        elif (self.P is None) and (self.electronic_state == ElectronicState.B):
            self.basis = Basis.CoupledΩ
        elif self.electronic_state == ElectronicState.B:
            self.basis = Basis.CoupledP
        elif self.electronic_state == ElectronicState.X:
            self.basis = Basis.Coupled
        else:
            self.basis = None
        self.v = v

    # equality testing
    def __eq__(self, other: object) -> bool:
        # return self.__hash__() == other.__hash__()
        if not isinstance(other, CoupledBasisState):
            return False
        else:
            return (
                self.F == other.F
                and self.mF == other.mF
                and self.I1 == other.I1
                and self.I2 == other.I2
                and self.F1 == other.F1
                and self.J == other.J
                and self.Omega == other.Omega
                and self.P == other.P
                and self.electronic_state == other.electronic_state
                and self.v == other.v
                and self.basis == other.basis
            )

    @overload
    def __matmul__(self, other: CoupledBasisState) -> Literal[0, 1]: ...

    @overload
    def __matmul__(self, other: UncoupledBasisState) -> UncoupledState: ...

    # inner product
    def __matmul__(self, other):
        if not isinstance(other, BasisState):
            raise TypeError(
                "can only matmul CoupledBasisState with CoupledBasisState or "
                f"UncoupledBasisState (not {type(other)})"
            )
        if other.isCoupled:
            if self == other:
                return 1
            else:
                return 0
        else:
            return UncoupledState([(1, other)]) @ self.transform_to_uncoupled()

    # superposition: addition
    def __add__(self, other: Self) -> CoupledState:
        if self == other:
            return CoupledState([(2, self)])
        elif isinstance(other, CoupledBasisState):
            if self.basis == other.basis:
                return CoupledState([(1, self), (1, other)])
            else:
                raise TypeError("can only add BasisStates with the same basis")
        else:
            raise TypeError(
                f"can only add CoupledBasisState (not {type(other)}) "
                "to CoupledBasisState"
            )

    # superposition: subtraction
    def __sub__(self, other: Self) -> CoupledState | Literal[0]:
        if self == other:
            return 0
        elif isinstance(other, CoupledBasisState):
            if self.basis == other.basis:
                return CoupledState([(1, self), (-1, other)])
            else:
                raise TypeError("can only add BasisStates withe same basis")
        else:
            raise TypeError(
                f"can only subtract CoupledBasisState (not {type(other)}) "
                "from CoupledBasisState"
            )

    # scalar product (psi * a)
    def __mul__(self, a: Union[complex, float, int]) -> CoupledState:
        return CoupledState([(a, self)])

    # scalar product (a * psi)
    def __rmul__(self, a: Union[float, complex, int]) -> CoupledState:
        return self * a

    def __hash__(self) -> int:
        if self.P is not None:
            P = self.P
        else:
            # parity 2 for the hash if the parity isn't defined
            P = 2
        ev = self.electronic_state.value if self.electronic_state is not None else 0
        quantum_numbers = (
            int(self.J),
            float(self.F1),
            int(self.F),
            int(self.mF),
            float(self.I1),
            float(self.I2),
            int(P),
            int(self.Omega),
            ev,
        )
        return int(hashlib.md5(json.dumps(quantum_numbers).encode()).hexdigest(), 16)

    def __repr__(self) -> str:
        return self.state_string()

    def state_string(self) -> str:
        F = sp.S(str(self.F), rational=True)
        mF = sp.S(str(self.mF), rational=True)
        F1 = sp.S(str(self.F1), rational=True)
        J = sp.S(str(self.J), rational=True)
        I1 = sp.S(str(self.I1), rational=True)
        I2 = sp.S(str(self.I2), rational=True)
        if self.P is not None:
            if self.P == 1:
                P = "+"
            elif self.P == -1:
                P = "-"
        Omega = self.Omega
        v = self.v

        string = f"J = {J}, F₁ = {F1}, F = {F}, mF = {mF}, I₁ = {I1}, I₂ = {I2}"

        if self.electronic_state is not None:
            string = f"{self.electronic_state.name}, {string}"
        if self.P is not None:
            string = f"{string}, P = {P}"
        if Omega is not None:
            string = f"{string}, Ω = {Omega}"
        if v is not None:
            string = f"{string}, v = {v}"
        return "|" + string + ">"

    def state_string_custom(self, to_print: List[str]) -> str:
        F = sp.S(str(self.F), rational=True)
        mF = sp.S(str(self.mF), rational=True)
        F1 = sp.S(str(self.F1), rational=True)
        J = sp.S(str(self.J), rational=True)
        I1 = sp.S(str(self.I1), rational=True)
        I2 = sp.S(str(self.I2), rational=True)
        if self.P is not None:
            if self.P == 1:
                P = "+"
            elif self.P == -1:
                P = "-"
        Omega = self.Omega
        v = self.v

        string = ""
        for val in to_print:
            if val == "J":
                string += f"J = {J}, "
            elif val == "F1":
                string += f"F₁ = {F1}, "
            elif val == "F":
                string += f"F = {F}, "
            elif val == "mF":
                string += f"mF = {mF}, "
            elif val == "I1":
                string += f"I₁ = {I1}, "
            elif val == "I2":
                string += f"I₂ = {I2}, "
            elif val == "electronic":
                if self.electronic_state is not None:
                    string = f"{self.electronic_state.name}, {string}"
            elif val == "P":
                if P is not None:
                    string += f"P = {P}, "
            elif val == "Ω":
                if Omega is not None:
                    string += f"Ω = {Omega}, "
            elif val == "v":
                if v is not None:
                    string += f"v = {v}, "
        string = string.strip(", ")
        return "|" + string + ">"

    def print_quantum_numbers(self, printing: bool = False) -> str:
        if printing:
            print(self.state_string())
        return self.state_string()

    # A method to transform from coupled to uncoupled basis
    def transform_to_uncoupled(self) -> UncoupledState:
        F = self.F
        mF = self.mF
        F1 = self.F1
        J = self.J
        I1 = self.I1
        I2 = self.I2
        electronic_state = self.electronic_state
        P = self.P
        Omega = self.Omega

        mF1s = np.arange(-F1, F1 + 1, 1)
        mJs = np.arange(-J, J + 1, 1)
        m1s = np.arange(-I1, I1 + 1, 1)
        m2s = np.arange(-I2, I2 + 1, 1)

        uncoupled_state = UncoupledState()

        for mF1 in mF1s:
            for mJ in mJs:
                for m1 in m1s:
                    for m2 in m2s:
                        amp = CGc(J, mJ, I1, m1, F1, mF1) * CGc(F1, mF1, I2, m2, F, mF)
                        basis_state = UncoupledBasisState(
                            J,
                            mJ,
                            I1,
                            m1,
                            I2,
                            m2,
                            P=P,
                            Omega=Omega,
                            electronic_state=electronic_state,
                        )
                        uncoupled_state = uncoupled_state + UncoupledState(
                            [(amp, basis_state)]
                        )

        return uncoupled_state.normalize()

    # Method for transforming parity eigenstate to Omega eigenstate basis
    def transform_to_omega_basis(self) -> CoupledState:
        F = self.F
        mF = self.mF
        F1 = self.F1
        J = self.J
        I1 = self.I1
        I2 = self.I2
        electronic_state = self.electronic_state
        P = self.P
        Omega = self.Omega

        assert self.basis is not None, "Unknown basis state, can't transform to Ω basis"
        assert (
            P is not None
        ), "Can't transform state to Omega basis if parity is not known"

        # Check that not already in omega basis
        if self.basis == Basis.CoupledP and self.electronic_state == ElectronicState.B:
            state_minus = 1 * CoupledBasisState(
                F,
                mF,
                F1,
                J,
                I1,
                I2,
                Omega=-1 * Omega,
                P=None,
                electronic_state=electronic_state,
                basis=Basis.CoupledΩ,
                v=self.v,
            )
            state_plus = 1 * CoupledBasisState(
                F,
                mF,
                F1,
                J,
                I1,
                I2,
                Omega=1 * Omega,
                P=None,
                electronic_state=electronic_state,
                basis=Basis.CoupledΩ,
                v=self.v,
            )
            state: CoupledState = (
                1 / np.sqrt(2) * (state_plus + P * (-1) ** (J) * state_minus)
            )
            return state

        elif self.basis == Basis.Coupled and self.electronic_state == ElectronicState.X:
            raise ValueError("Cannot transform X state to Omega basis")
        else:
            raise ValueError("Cannot transform to Omega basis")

    def transform_to_parity_basis(self) -> CoupledState:
        """
        Transforms self from Omega eigenstate basis (i.e. signed Omega) to
        parity eigenstate basis (unsigned Omega, P is good quantum number).
        Doing this is only defined for electronic state B.
        """
        F = self.F
        mF = self.mF
        F1 = self.F1
        J = self.J
        I1 = self.I1
        I2 = self.I2
        electronic_state = self.electronic_state
        P = self.P
        Omega = self.Omega
        S = 0

        # Check that not already in parity basis

        if (self.basis == Basis.CoupledΩ) or (
            P is None and not electronic_state == ElectronicState.X
        ):
            if np.sign(Omega) == 1:
                state = (
                    1
                    / np.sqrt(2)
                    * (
                        1
                        * CoupledBasisState(
                            F,
                            mF,
                            F1,
                            J,
                            I1,
                            I2,
                            Omega=np.abs(Omega),
                            P=+1,
                            electronic_state=electronic_state,
                        )
                        + 1
                        * CoupledBasisState(
                            F,
                            mF,
                            F1,
                            J,
                            I1,
                            I2,
                            Omega=np.abs(Omega),
                            P=-1,
                            electronic_state=electronic_state,
                        )
                    )
                )

            elif np.sign(Omega) == -1:
                state = (
                    1
                    / np.sqrt(2)
                    * (-1) ** (J - S)
                    * (
                        1
                        * CoupledBasisState(
                            F,
                            mF,
                            F1,
                            J,
                            I1,
                            I2,
                            Omega=np.abs(Omega),
                            P=+1,
                            electronic_state=electronic_state,
                        )
                        - 1
                        * CoupledBasisState(
                            F,
                            mF,
                            F1,
                            J,
                            I1,
                            I2,
                            Omega=np.abs(Omega),
                            P=-1,
                            electronic_state=electronic_state,
                        )
                    )
                )
        else:
            state = 1 * self

        return state


# Class for uncoupled basis states
class UncoupledBasisState(BasisState):
    # constructor
    def __init__(
        self,
        J: int,
        mJ: int,
        I1: float,
        m1: float,
        I2: float,
        m2: float,
        Omega: Optional[int] = None,
        P: Optional[int] = None,
        electronic_state: Optional[ElectronicState] = None,
        energy: Optional[float] = None,
        basis: Optional[Basis] = None,
    ):
        self.J, self.mJ = J, mJ
        self.I1, self.m1 = I1, m1
        self.I2, self.m2 = I2, m2
        self.Omega = Omega
        if P is not None:
            self.P = P
        else:
            raise AssertionError("need to supply parity P")
        self.electronic_state = electronic_state
        self.isCoupled = False
        self.isUncoupled = True

        if basis is not None:
            self.basis = basis
        else:
            self.basis = Basis.Uncoupled

        self.energy = energy

    # equality testing
    def __eq__(self, other: object) -> bool:
        if not isinstance(other, UncoupledBasisState):
            return False
        else:
            return (
                self.J == other.J
                and self.mJ == other.mJ
                and self.I1 == other.I1
                and self.I2 == other.I2
                and self.m1 == other.m1
                and self.m2 == other.m2
                and self.Omega == other.Omega
                and self.P == other.P
                and self.electronic_state == other.electronic_state
                and self.basis == other.basis
            )

    @overload
    def __matmul__(self, other: UncoupledBasisState) -> Literal[1, 0]: ...

    @overload
    def __matmul__(self, other: CoupledBasisState) -> UncoupledState: ...

    # inner product
    def __matmul__(self, other):
        if other.isUncoupled:
            if self == other:
                return 1
            else:
                return 0
        elif isinstance(other, CoupledBasisState):
            return UncoupledState([(1, self)]) @ other.transform_to_uncoupled()
        else:
            raise TypeError(
                "can only multiply UncoupledBasisState with UncoupledBasisState or "
                f"CoupledBasisState (not {type(other)}"
            )

    # superposition: addition
    def __add__(self, other: UncoupledBasisState) -> UncoupledState:
        if self == other:
            return UncoupledState([(2, self)])
        elif isinstance(other, UncoupledBasisState):
            return UncoupledState([(1, self), (1, other)])
        else:
            raise TypeError(
                f"can only add UncoupledBasisState (not {type(other)}) "
                "to UncoupledBasisState"
            )

    # superposition: subtraction
    def __sub__(self, other: UncoupledBasisState) -> Union[int, UncoupledState]:
        if self == other:
            return 0
        elif isinstance(other, UncoupledBasisState):
            return UncoupledState([(1, self), (-1, other)])
        else:
            raise TypeError(
                f"can only subtract UncoupledBasisState (not {type(other)}) "
                "from UncoupledBasisState"
            )

    # scalar product (psi * a)
    def __mul__(self, a: Union[float, complex, int]) -> UncoupledState:
        return UncoupledState([(a, self)])

    # scalar product (a * psi)
    def __rmul__(self, a: Union[float, complex, int]) -> UncoupledState:
        return self * a

    def __hash__(self):
        quantum_numbers = (
            int(self.J),
            int(self.mJ),
            float(self.I1),
            float(self.m1),
            float(self.I2),
            float(self.m2),
            int(self.P),
            int(self.Omega),
            self.electronic_state.value,
        )
        return int(hashlib.md5(json.dumps(quantum_numbers).encode()).hexdigest(), 16)

    def __repr__(self) -> str:
        return self.state_string()

    def state_string(self) -> str:
        J, mJ = sp.S(str(self.J), rational=True), sp.S(str(self.mJ), rational=True)
        I1 = sp.S(str(self.I1), rational=True)
        m1 = sp.S(str(self.m1), rational=True)
        I2 = sp.S(str(self.I2), rational=True)
        m2 = sp.S(str(self.m2), rational=True)
        if self.P == 1:
            P = "+"
        elif self.P == -1:
            P = "-"
        Omega = self.Omega

        string = f"J = {J}, mJ = {mJ}, I₁ = {I1}, m₁ = {m1}, I₂ = {I2}, m₂ = {m2}"

        if self.electronic_state is not None:
            string = f"{self.electronic_state.name}, {string}"
        if self.P is not None:
            string = f"{string}, P = {P}"
        if Omega is not None:
            string = f"{string}, Ω = {Omega}"
        return "|" + string + ">"

    def print_quantum_numbers(self, printing: bool = False) -> str:
        if printing:
            print(self.state_string())
        return self.state_string()

    # Method for converting to coupled basis
    def transform_to_coupled(self) -> CoupledState:
        # Determine quantum numbers
        J = self.J
        mJ = self.mJ
        I1 = self.I1
        m1 = self.m1
        I2 = self.I1
        m2 = self.m2
        Omega = self.Omega
        electronic_state = self.electronic_state
        P = self.P
        Omega = 0 if self.Omega is None else self.Omega

        # Determine what mF has to be
        mF = int(mJ + m1 + m2)

        uncoupled_state = self

        data = []

        # Loop over possible values of F1, F and m_F
        for F1 in np.arange(J - I1, J + I1 + 1):
            for F in np.arange(F1 - I2, F1 + I2 + 1):
                if np.abs(mF) <= F:
                    coupled_state = CoupledBasisState(
                        F,
                        mF,
                        F1,
                        J,
                        I1,
                        I2,
                        Omega=Omega,
                        P=P,
                        electronic_state=electronic_state,
                    )
                    amp = uncoupled_state @ coupled_state
                    data.append((amp, coupled_state))

        return CoupledState(data)

    # Method for transforming parity eigenstate to Omega eigenstate basis
    def transform_to_omega_basis(self) -> UncoupledState:
        # Determine quantum numbers
        J = self.J
        mJ = self.mJ
        I1 = self.I1
        m1 = self.m1
        I2 = self.I1
        m2 = self.m2
        Omega = self.Omega
        electronic_state = self.electronic_state
        P = self.P
        Omega = 0 if self.Omega is None else self.Omega

        # Check that not already in omega basis
        if P is not None and not electronic_state == "X":
            state_minus = 1 * UncoupledBasisState(
                J,
                mJ,
                I1,
                m1,
                I2,
                m2,
                P=P,
                Omega=-1 * Omega,
                electronic_state=electronic_state,
            )
            state_plus = 1 * UncoupledBasisState(
                J,
                mJ,
                I1,
                m1,
                I2,
                m2,
                P=P,
                Omega=Omega,
                electronic_state=electronic_state,
            )

            state = 1 / np.sqrt(2) * (state_plus + P * (-1) ** (J - 1) * state_minus)
        else:
            state = 1 * self

        return state


S = TypeVar("S", bound=BasisState)


class State(Generic[S]):
    def __init__(
        self,
        data: Sequence[
            Tuple[
                Union[int, float, complex],
                S,
            ]
        ],
        remove_zero_amp_cpts: bool = True,
    ):
        # remove components with zero amplitudes
        if remove_zero_amp_cpts:
            self.data = [(amp, cpt) for amp, cpt in data if amp != 0]
        else:
            self.data = list(data)

        # for iteration over the State
        self.index = len(self.data)

    def _create_new_instance(
        self,
        data: Sequence[
            Tuple[
                Union[int, float, complex],
                S,
            ]
        ],
        remove_zero_amp_cpts: bool = True,
    ) -> Self:
        return self.__class__(data, remove_zero_amp_cpts)

    # superposition: addition
    # (highly inefficient and ugly but should work)
    def __add__(self, other: Self) -> Self:
        data: List[Tuple[Union[float, complex], S]] = []
        # add components that are in self but not in other
        for amp1, cpt1 in self.data:
            only_in_self = True
            for amp2, cpt2 in other.data:
                if cpt2 == cpt1:
                    only_in_self = False
                    break
            if only_in_self:
                data.append((amp1, cpt1))
        # add components that are in other but not in self
        for amp1, cpt1 in other.data:
            only_in_other = True
            for amp2, cpt2 in self.data:
                if cpt2 == cpt1:
                    only_in_other = False
                    break
            if only_in_other:
                data.append((amp1, cpt1))
        # add components that are both in self and in other
        for amp1, cpt1 in self.data:
            for amp2, cpt2 in other.data:
                if cpt2 == cpt1:
                    data.append((amp1 + amp2, cpt1))
                    break
        return self._create_new_instance(data)

    # superposition: subtraction
    def __sub__(self, other: Self) -> Self:
        return self + -1 * other

    # scalar product (psi * a)
    def __mul__(self, a: Union[float, complex, int]) -> Self:
        return self._create_new_instance([(a * amp, psi) for amp, psi in self.data])

    # scalar product (a * psi)
    def __rmul__(self, a: Union[float, complex, int]) -> Self:
        return self * a

    # scalar division (psi / a)
    def __truediv__(self, a: Union[float, complex, int]) -> Self:
        return self * (1 / a)

    # negation
    def __neg__(self) -> Self:
        return -1 * self

    # inner product
    def __matmul__(self, other: Self) -> Union[int, float, complex]:
        result = 0
        for amp1, psi1 in self:
            for amp2, psi2 in other:
                result += amp1.conjugate() * amp2 * (psi1 @ psi2)
        return result

    # iterator methods
    def __iter__(self):
        return ((amp, state) for amp, state in self.data)

    def __next__(self):
        if self.index == 0:
            raise StopIteration
        self.index -= 1
        return self.data[self.index]

    def __eq__(self, other: object) -> bool:
        """
        __eq__ method
        iterates through all amplitudes and states, can be slow

        Args:
            other (object): object to compare to

        Returns:
            bool: True if equal, False if not
        """
        if not isinstance(other, self.__class__):
            return False
        else:
            S1 = self.order_by_amp()
            S2 = other.order_by_amp()
            for (a1, s1), (a2, s2) in zip(S1, S2):
                if a1 != a2:
                    return False
                elif s1 != s2:
                    return False
            return True

    # direct access to a component
    def __getitem__(self, i: int) -> Tuple[Union[complex, float, int], BasisState]:
        return self.data[i]

    def __repr__(self) -> str:
        ordered = self.order_by_amp()
        idx = 0
        string = ""
        amp_max = np.max(np.abs(list(zip(*ordered))[0]))
        for amp, state in ordered:
            if np.abs(amp) < amp_max * 1e-3:
                continue
            string += f"{amp:.2f} x {state}"
            idx += 1
            if (idx > 5) or (idx == len(ordered.data)):
                break
            string += "\n"
        if idx == 0:
            return ""
        else:
            return string

    def find_largest_component(self) -> S:
        # Order the state by amplitude
        state = self.order_by_amp()

        return state.data[0][1]

    @property
    def largest(self) -> S:
        return self.find_largest_component()

    def normalize(self) -> Self:
        data = []
        N = np.sqrt(self @ self)
        for amp, basis_state in self.data:
            data.append((amp / N, basis_state))

        return self._create_new_instance(data)

    # Method that removes components that are smaller than tolerance from the state
    def remove_small_components(self, tol: float = 1e-3) -> Self:
        purged_data = []
        for amp, basis_state in self.data:
            if np.abs(amp) > tol:
                purged_data.append((amp, basis_state))

        return self._create_new_instance(purged_data)

    # Method for ordering states in descending order of amp^2
    def order_by_amp(self) -> Self:
        data = self.data
        amp_array = np.zeros(len(data))

        # Make an numpy array of the amplitudes
        for i, d in enumerate(data):
            amp_array[i] = np.abs((data[i][0])) ** 2

        # Find ordering of array in descending order
        index = np.argsort(-1 * amp_array)

        # Reorder data
        reordered_data = data
        reordered_data = [reordered_data[i] for i in index]

        return self._create_new_instance(reordered_data)

    # Method for printing largest component basis states
    def print_largest_components(self, n: int = 1) -> str:
        # Order the state by amplitude
        state = self.order_by_amp()

        # Initialize an empty string
        string = ""

        for i in range(0, n):
            basis_state = state.data[i][1]
            basis_state.print_quantum_numbers()

        return string

    def state_vector(
        self,
        QN: Sequence[S],
    ) -> npt.NDArray[np.complex128]:
        state_vector = [1 * state @ self for state in QN]
        return np.array(state_vector, dtype=complex)

    # Method that generates a density matrix from state
    def density_matrix(
        self,
        QN: Sequence[S],
    ) -> npt.NDArray[np.complex128]:
        # Get state vector
        state_vec = self.state_vector(QN)

        # Generate density matrix from state vector
        density_matrix = np.tensordot(state_vec.conj(), state_vec, axes=0)

        return density_matrix

    # Method for transforming all basis states to omega basis
    def transform_to_omega_basis(self) -> Self:
        state = self._create_new_instance(data=[])
        for amp, basis_state in self.data:
            state += amp * basis_state.transform_to_omega_basis()

        return state

    # Method for transforming all basis states to parity basis
    def transform_to_parity_basis(self) -> Self:
        state = self._create_new_instance(data=[])
        for amp, basis_state in self.data:
            state += amp * basis_state.transform_to_parity_basis()

        return state


class CoupledState(State):
    def __init__(
        self,
        data: Sequence[Tuple[Union[int, float, complex], CoupledBasisState]] = [],
        remove_zero_amp_cpts: bool = True,
    ):
        super().__init__(data=data, remove_zero_amp_cpts=remove_zero_amp_cpts)

    def _create_new_instance(
        self,
        data: Sequence[Tuple[Union[int, float, complex], CoupledBasisState]],
        remove_zero_amp_cpts: bool = True,
    ):
        return CoupledState(data, remove_zero_amp_cpts)

    def find_largest_component(self) -> CoupledBasisState:
        return super().find_largest_component()

    @property
    def largest(self) -> CoupledBasisState:
        return super().largest

    # Function that returns state vector in given basis
    def state_vector(
        self,
        QN: Union[
            Sequence[CoupledBasisState],
            Sequence[CoupledState],
        ],
    ) -> npt.NDArray[np.complex128]:
        return super().state_vector(QN)

    # Method that generates a density matrix from state
    def density_matrix(
        self, QN: Union[Sequence[CoupledBasisState], Sequence[CoupledState]]
    ) -> npt.NDArray[np.complex128]:
        return super().density_matrix(QN)

    # Method for converting the state into the uncoupled basis
    def transform_to_uncoupled(self) -> UncoupledState:
        # Loop over the basis states, check if they are already in uncoupled
        # basis and if not convert to uncoupled basis, output state in new basis
        state_in_uncoupled_basis = UncoupledState()

        for amp, basis_state in self.data:
            if basis_state.isUncoupled:
                state_in_uncoupled_basis += UncoupledState([(amp, basis_state)])
            if basis_state.isCoupled:
                state_in_uncoupled_basis += amp * basis_state.transform_to_uncoupled()

        return state_in_uncoupled_basis


class UncoupledState(State):
    def __init__(
        self,
        data: Sequence[Tuple[Union[int, float, complex], UncoupledBasisState]] = [],
        remove_zero_amp_cpts: bool = True,
    ):
        super().__init__(data=data, remove_zero_amp_cpts=remove_zero_amp_cpts)

    def _create_new_instance(
        self,
        data: Sequence[Tuple[int | float | complex, UncoupledBasisState]],
        remove_zero_amp_cpts: bool = True,
    ) -> UncoupledState:
        return UncoupledState(data, remove_zero_amp_cpts)

    def find_largest_component(self) -> UncoupledBasisState:
        return super().find_largest_component()

    @property
    def largest(self) -> UncoupledBasisState:
        return super().largest

    # Function that returns state vector in given basis
    def state_vector(
        self,
        QN: Union[
            Sequence[UncoupledBasisState],
            Sequence[UncoupledState],
        ],
    ) -> npt.NDArray[np.complex128]:
        return super().state_vector(QN)

    # Method that generates a density matrix from state
    def density_matrix(
        self, QN: Union[Sequence[UncoupledBasisState], Sequence[UncoupledState]]
    ) -> npt.NDArray[np.complex128]:
        return super().density_matrix(QN)

    # Method for converting the state into the coupled basis
    def transform_to_coupled(self) -> CoupledState:
        # Loop over the basis states, check if they are already in uncoupled
        # basis and if not convert to coupled basis, output state in new basis
        state_in_coupled_basis = CoupledState()

        for amp, basis_state in self.data:
            if basis_state.isCoupled:
                state_in_coupled_basis += CoupledState([(amp, basis_state)])
            if basis_state.isUncoupled:
                state_in_coupled_basis += amp * basis_state.transform_to_coupled()

        return state_in_coupled_basis


def BasisStates_from_State(
    states: Union[Sequence[State], npt.NDArray[Any]],
) -> npt.NDArray[Any]:
    unique = []
    for state in states:
        for amp, basisstate in state:
            if basisstate not in unique:
                unique.append(basisstate)
    return np.array(unique)


@overload
def basisstate_to_state_list(
    states: Sequence[CoupledBasisState],
) -> List[CoupledState]: ...


@overload
def basisstate_to_state_list(
    states: Sequence[UncoupledBasisState],
) -> List[UncoupledState]: ...


def basisstate_to_state_list(states):
    return [1 * bs for bs in states]



from centrex_tlf.states import UncoupledBasisState, UncoupledState

from .constants import HamiltonianConstants
from .quantum_operators import J2

__all__ = ["Hrot"]

########################################################
# Rotational Term
########################################################


def Hrot(
    psi: UncoupledBasisState, coefficients: HamiltonianConstants
) -> UncoupledState:
    return coefficients.B_rot * J2(psi)



from functools import lru_cache
from typing import Tuple

import numpy as np
import numpy.typing as npt

from centrex_tlf import states

from .wigner import sixj_f, threej_f


def generate_ED_ME_mixed_state(
    bra: states.CoupledState,
    ket: states.CoupledState,
    pol_vec: npt.NDArray[np.complex_] = np.array([1.0, 1.0, 1.0], dtype=np.complex_),
    reduced: bool = False,
    normalize_pol: bool = True,
) -> complex:
    ME = 0j

    # Transform to Omega basis if required. For the X state the basis is Coupled and
    # doesn't require to be transformed to the Omega basis, since Omega is 0.
    if bra.largest.basis is states.Basis.CoupledP:
        bra = bra.transform_to_omega_basis()
    if ket.largest.basis is states.Basis.CoupledP:
        ket = ket.transform_to_omega_basis()

    if normalize_pol:
        pol_vec /= np.linalg.norm(pol_vec)

    for amp_bra, basis_bra in bra.data:
        for amp_ket, basis_ket in ket.data:
            ME += (
                amp_bra.conjugate()
                * amp_ket
                * ED_ME_coupled(
                    basis_bra, basis_ket, pol_vec=tuple(pol_vec), rme_only=reduced
                )
            )

    return ME


@lru_cache(maxsize=int(1e6))
def ED_ME_coupled(
    bra: states.CoupledBasisState,
    ket: states.CoupledBasisState,
    pol_vec: Tuple[complex, complex, complex] = (1.0 + 0j, 1.0 + 0j, 1.0 + 0j),
    rme_only: bool = False,
) -> complex:
    """calculate electric dipole matrix elements between coupled basis states

    Args:
        bra (CoupledBasisState): coupled basis state object
        ket (CoupledBasisState): coupled basis state object
        pol_vec (Tuple[float, float, float]): polarization vector.
                                        Defaults to np.array([1,1,1]).
        rme_only (bool, optional): set True to return only reduced matrix
                                    element, otherwise angular component is
                                    included. Defaults to False.

    Returns:
        complex: electric dipole matrix element between bra en ket
    """

    # find quantum numbers for ground state
    F = bra.F
    mF = bra.mF
    J = bra.J
    F1 = bra.F1
    I1 = bra.I1
    I2 = bra.I2
    Omega = bra.Omega

    # find quantum numbers for excited state
    Fp = ket.F
    mFp = ket.mF
    Jp = ket.J
    F1p = ket.F1
    Omegap = ket.Omega

    # calculate the reduced matrix element
    q = Omega - Omegap
    ME = (
        (-1) ** (F1 + J + Fp + F1p + I1 + I2)
        * np.sqrt((2 * F + 1) * (2 * Fp + 1) * (2 * F1p + 1) * (2 * F1 + 1))
        * sixj_f(F1p, Fp, I2, F, F1, 1)
        * sixj_f(Jp, F1p, I1, F1, J, 1)
        * (-1) ** (J - Omega)
        * np.sqrt((2 * J + 1) * (2 * Jp + 1))
        * threej_f(J, 1, Jp, -Omega, q, Omegap)
        * float(np.abs(q) < 2)
    )

    # if we want the complete matrix element, calculate angular part
    if not rme_only:
        # calculate elements of the polarization vector in spherical basis
        p_vec = {}
        p_vec[-1] = -1 / np.sqrt(2) * (pol_vec[0] + 1j * pol_vec[1])
        p_vec[0] = pol_vec[2]
        p_vec[1] = +1 / np.sqrt(2) * (pol_vec[0] - 1j * pol_vec[1])

        # calculate the value of p that connects the states
        p = mF - mFp
        p = p * int(np.abs(p) <= 1)
        # multiply RME by the angular part
        ME = (
            ME
            * (-1) ** (F - mF)
            * threej_f(F, 1, Fp, -mF, p, mFp)
            * p_vec[p]
            * int(np.abs(p) <= 1)
        )

    # return the matrix element
    return ME



from dataclasses import dataclass

import numpy as np
import numpy.typing as npt

__all__ = [
    "polarization_X",
    "polarization_Y",
    "polarization_Z",
    "polarization_σp",
    "polarization_σm",
]


@dataclass
class Polarization:
    vector: npt.NDArray[np.complex_]
    name: str

    def __repr__(self) -> str:
        return f"Polarization({self.name})"


polarization_X = Polarization(np.array([1, 0, 0], dtype=np.complex128), "X")
polarization_Y = Polarization(np.array([0, 1, 0], dtype=np.complex128), "Y")
polarization_Z = Polarization(np.array([0, 0, 1], dtype=np.complex128), "Z")
polarization_σp = Polarization(
    np.array([1 / np.sqrt(2), 1j / np.sqrt(2), 0], dtype=np.complex128), "σp"
)
polarization_σm = Polarization(
    np.array([1 / np.sqrt(2), -1j / np.sqrt(2), 0], dtype=np.complex128), "σm"
)



from functools import lru_cache

import numpy as np

from centrex_tlf.states import CoupledBasisState, CoupledState

from ..constants import BConstants
from ..wigner import sixj_f


@lru_cache(maxsize=int(1e6))
def H_c_Tl(psi: CoupledBasisState, constants: BConstants) -> CoupledState:
    """
    Calculates the effect of the c1 term on the input basis state
    """
    # Find the quantum numbers of the input state
    Jp = psi.J
    I1p = psi.I1
    I2p = psi.I2
    F1p = psi.F1
    Fp = psi.F
    mFp = psi.mF
    Omegap = psi.Omega

    # J, I1, I2, F and mF are the same for both states
    J = Jp
    I1 = I1p
    I2 = I2p
    F = Fp
    F1 = F1p
    mF = mFp

    # Omega also doesn't change
    Omega = Omegap

    # Calculate matrix element
    amp = (
        constants.c_Tl
        * (-1) ** (I1 + F1 + J)
        * sixj_f(Jp, I1, F1, I1, J, 1)
        * np.sqrt(I1 * (I1 + 1) * (2 * I1 + 1) * J * (J + 1) * (2 * J + 1))
    )

    basis_state = CoupledBasisState(
        F,
        mF,
        F1,
        J,
        I1,
        I2,
        Omega=Omega,
        electronic_state=psi.electronic_state,
        P=psi.P,
    )

    return CoupledState([(amp, basis_state)])



from functools import lru_cache

import numpy as np

from centrex_tlf.states import CoupledBasisState, CoupledState

from ..constants import BConstants
from ..wigner import sixj_f, threej_f


@lru_cache(maxsize=int(1e6))
def H_mhf_Tl(psi: CoupledBasisState, constants: BConstants) -> CoupledState:
    """
    Operator for magnetic hyperfine term for Tl nucleus.
    """
    # Find the quantum numbers of the input state
    Jp = psi.J
    I1p = psi.I1
    I2p = psi.I2
    F1p = psi.F1
    Fp = psi.F
    mFp = psi.mF
    Omegap = psi.Omega

    # I1, I2, F1, F and mF are the same for both states
    I1 = I1p
    I2 = I2p
    F1 = F1p
    F = Fp
    mF = mFp

    # Omega also doesn't change
    Omega = Omegap

    # Initialize container for storing states and matrix elements
    data = []

    # Loop over the possible values of quantum numbers for which the matrix element can
    # be non-zero
    # Need J = Jp+1 ... |Jp-1|
    for J in np.arange(np.abs(Jp - 1), Jp + 2):
        # Calculate matrix element
        amp = (
            Omega
            * constants.h1_Tl
            * (-1) ** (J + Jp + F1 + I1 - Omega)
            * sixj_f(I1, Jp, F1, J, I1, 1)
            * threej_f(J, 1, Jp, -Omega, 0, Omegap)
            * np.sqrt((2 * J + 1) * (2 * Jp + 1) * I1 * (I1 + 1) * (2 * I1 + 1))
        )

        basis_state = CoupledBasisState(
            F,
            mF,
            F1,
            J,
            I1,
            I2,
            Omega=Omega,
            electronic_state=psi.electronic_state,
            P=psi.P,
        )
        if amp != 0:
            data.append((amp, basis_state))

    return CoupledState(data)


@lru_cache(maxsize=int(1e6))
def H_mhf_F(psi: CoupledBasisState, constants: BConstants) -> CoupledState:
    """
    Operator for magnetic hyperfine term for F nucleus.
    """
    # Find the quantum numbers of the input state
    Jp = psi.J
    I1p = psi.I1
    I2p = psi.I2
    F1p = psi.F1
    Fp = psi.F
    mFp = psi.mF
    Omegap = psi.Omega

    # I1, I2, F and mF are the same for both states
    I1 = I1p
    I2 = I2p
    F = Fp
    mF = mFp

    # Omega also doesn't change
    Omega = Omegap

    # Initialize container for storing states and matrix elements
    data = []

    # Loop over the possible values of quantum numbers for which the matrix element can
    # be non-zero
    # Need J = Jp+1 ... |Jp-1|
    for J in np.arange(np.abs(Jp - 1), Jp + 2):
        # F1 can be J +/- 1/2
        for F1 in np.arange(np.abs(J - 1 / 2), J + 3 / 2):
            # Calculate matrix element
            amp = (
                Omega
                * constants.h1_F
                * (-1) ** (2 * F1p + F + 2 * J + I1 + I2 - Omega + 1)
                * sixj_f(I2, F1p, F, F1, I2, 1)
                * sixj_f(Jp, F1p, I1, F1, J, 1)
                * threej_f(J, 1, Jp, -Omega, 0, Omegap)
                * np.sqrt(
                    (2 * F1 + 1)
                    * (2 * F1p + 1)
                    * (2 * J + 1)
                    * (2 * Jp + 1)
                    * I2
                    * (I2 + 1)
                    * (2 * I2 + 1)
                )
            )

            basis_state = CoupledBasisState(
                F,
                mF,
                F1,
                J,
                I1,
                I2,
                Omega=Omega,
                electronic_state=psi.electronic_state,
                P=psi.P,
            )
            if amp != 0:
                data.append((amp, basis_state))

    return CoupledState(data)



from functools import lru_cache

import numpy as np

from centrex_tlf.states import CoupledBasisState, CoupledState

from ..constants import BConstants
from ..wigner import sixj_f, threej_f


@lru_cache(maxsize=int(1e6))
def d_p(psi: CoupledBasisState, p: int, constants: BConstants) -> CoupledState:
    """
    Operates on psi using the pth spherical tensor component of the
    dipole operator.
    """
    # Find the quantum numbers of the input state
    Jp = psi.J
    I1p = psi.I1
    I2p = psi.I2
    F1p = psi.F1
    Fp = psi.F
    mFp = psi.mF
    Omegap = psi.Omega

    # I1, I2 are the same for both states
    I1 = I1p
    I2 = I2p

    # Value of mF changes by p
    mF = mFp + p

    # Omega doesn't change
    Omega = Omegap

    # Initialize container for storing states and matrix elements
    data = []

    # Loop over possible values of Jprime
    for J in np.arange(np.abs(Jp - 1), Jp + 2):
        # Loop over possible values of F1
        for F1 in np.arange(np.abs(J - I1), J + I1 + 1):
            # Loop over possible values of F
            for F in np.arange(np.abs(F1 - I2), F1 + I2 + 1):
                amp = (
                    constants.μ_E
                    * (-1) ** (F + Fp + F1 + F1p + I1 + I2 - Omega - mF)
                    * np.sqrt(
                        (2 * F + 1)
                        * (2 * Fp + 1)
                        * (2 * F1 + 1)
                        * (2 * F1p + 1)
                        * (2 * J + 1)
                        * (2 * Jp + 1)
                    )
                    * threej_f(F, 1, Fp, -mF, p, mFp)
                    * threej_f(J, 1, Jp, -Omega, 0, Omegap)
                    * sixj_f(F1p, Fp, I2, F, F1, 1)
                    * sixj_f(Jp, F1p, I1, F1, J, 1)
                )

                basis_state = CoupledBasisState(
                    F,
                    mF,
                    F1,
                    J,
                    I1,
                    I2,
                    Omega=Omega,
                    electronic_state=psi.electronic_state,
                    P=psi.P,
                )
                if amp != 0:
                    data.append((amp, basis_state))

    return CoupledState(data)


@lru_cache(maxsize=int(1e6))
def HSx(psi: CoupledBasisState, constants: BConstants) -> CoupledState:
    """
    Stark Hamiltonian operator for x-component of electric field
    """
    return -(d_p(psi, -1, constants) - d_p(psi, +1, constants)) / np.sqrt(2)


@lru_cache(maxsize=int(1e6))
def HSy(psi: CoupledBasisState, constants: BConstants) -> CoupledState:
    """
    Stark Hamiltonian operator for y-component of electric field
    """
    return -1j * (d_p(psi, -1, constants) + d_p(psi, +1, constants)) / np.sqrt(2)


@lru_cache(maxsize=int(1e6))
def HSz(psi: CoupledBasisState, constants: BConstants) -> CoupledState:
    """
    Stark Hamiltonian for z-component of electric field
    """
    return -d_p(psi, 0, constants)



from functools import lru_cache

import numpy as np

from centrex_tlf.states import UncoupledBasisState, UncoupledState, parity_X

from .constants import XConstants
from .general_uncoupled import Hrot
from .quantum_operators import (
    I1m,
    I1p,
    I1x,
    I1y,
    I1z,
    I2m,
    I2p,
    I2x,
    I2y,
    I2z,
    Jm,
    Jp,
    Jx,
    Jy,
    Jz,
    com,
)

__all__ = [
    "Hc1",
    "Hc2",
    "Hc4",
    "Hc3a",
    "Hc3b",
    "Hc3c",
    "Hc3",
    "Hff",
    "HZx",
    "HZy",
    "HZz",
    "HSx",
    "HSy",
    "HSz",
    "R1p",
]

########################################################
# Terms with angular momentum dot products
########################################################


def Hc1(psi: UncoupledBasisState, coefficients: XConstants) -> UncoupledState:
    return coefficients.c1 * (
        com(I1z, Jz, psi, coefficients)
        + (1 / 2) * (com(I1p, Jm, psi, coefficients) + com(I1m, Jp, psi, coefficients))
    )


def Hc2(psi: UncoupledBasisState, coefficients: XConstants) -> UncoupledState:
    return coefficients.c2 * (
        com(I2z, Jz, psi, coefficients)
        + (1 / 2) * (com(I2p, Jm, psi, coefficients) + com(I2m, Jp, psi, coefficients))
    )


def Hc4(psi: UncoupledBasisState, coefficients: XConstants) -> UncoupledState:
    return coefficients.c4 * (
        com(I1z, I2z, psi, coefficients)
        + (1 / 2)
        * (com(I1p, I2m, psi, coefficients) + com(I1m, I2p, psi, coefficients))
    )


def Hc3a(psi: UncoupledBasisState, coefficients: XConstants) -> UncoupledState:
    return (
        15
        * coefficients.c3
        / coefficients.c1
        / coefficients.c2
        * com(Hc1, Hc2, psi, coefficients)
        / ((2 * psi.J + 3) * (2 * psi.J - 1))
    )


def Hc3b(psi: UncoupledBasisState, coefficients: XConstants) -> UncoupledState:
    return (
        15
        * coefficients.c3
        / coefficients.c2
        / coefficients.c1
        * com(Hc2, Hc1, psi, coefficients)
        / ((2 * psi.J + 3) * (2 * psi.J - 1))
    )


def Hc3c(psi: UncoupledBasisState, coefficients: XConstants) -> UncoupledState:
    return (
        -10
        * coefficients.c3
        / coefficients.c4
        / coefficients.B_rot
        * com(Hc4, Hrot, psi, coefficients)
        / ((2 * psi.J + 3) * (2 * psi.J - 1))
    )


def Hc3(psi: UncoupledBasisState, coefficients: XConstants) -> UncoupledState:
    return Hc3a(psi, coefficients) + Hc3b(psi, coefficients) + Hc3c(psi, coefficients)


########################################################
# Field free X state Hamiltonian
########################################################


def Hff(psi: UncoupledBasisState, coefficients: XConstants) -> UncoupledState:
    return (
        Hrot(psi, coefficients)
        + Hc1(psi, coefficients)
        + Hc2(psi, coefficients)
        + Hc3a(psi, coefficients)
        + Hc3b(psi, coefficients)
        + Hc3c(psi, coefficients)
        + Hc4(psi, coefficients)
    )


########################################################
# Zeeman X state
########################################################


@lru_cache(maxsize=int(1e6))
def HZx(psi: UncoupledBasisState, coefficients: XConstants) -> UncoupledState:
    if psi.J != 0:
        return (
            -coefficients.μ_J / psi.J * Jx(psi)
            - coefficients.μ_Tl / psi.I1 * I1x(psi)
            - coefficients.μ_F / psi.I2 * I2x(psi)
        )
    else:
        return -coefficients.μ_Tl / psi.I1 * I1x(psi) - coefficients.μ_F / psi.I2 * I2x(
            psi
        )


@lru_cache(maxsize=int(1e6))
def HZy(psi: UncoupledBasisState, coefficients: XConstants) -> UncoupledState:
    if psi.J != 0:
        return (
            -coefficients.μ_J / psi.J * Jy(psi)
            - coefficients.μ_Tl / psi.I1 * I1y(psi)
            - coefficients.μ_F / psi.I2 * I2y(psi)
        )
    else:
        return -coefficients.μ_Tl / psi.I1 * I1y(psi) - coefficients.μ_F / psi.I2 * I2y(
            psi
        )


@lru_cache(maxsize=int(1e6))
def HZz(psi: UncoupledBasisState, coefficients: XConstants) -> UncoupledState:
    if psi.J != 0:
        return (
            -coefficients.μ_J / psi.J * Jz(psi)
            - coefficients.μ_Tl / psi.I1 * I1z(psi)
            - coefficients.μ_F / psi.I2 * I2z(psi)
        )
    else:
        return -coefficients.μ_Tl / psi.I1 * I1z(psi) - coefficients.μ_F / psi.I2 * I2z(
            psi
        )


########################################################
# Stark Hamiltonian
########################################################


@lru_cache(maxsize=int(1e6))
def HSx(psi: UncoupledBasisState, coefficients: XConstants) -> UncoupledState:
    return (
        -coefficients.D_TlF
        * (R1m(psi, coefficients) - R1p(psi, coefficients))
        / np.sqrt(2)
    )


@lru_cache(maxsize=int(1e6))
def HSy(psi: UncoupledBasisState, coefficients: XConstants) -> UncoupledState:
    return (
        -coefficients.D_TlF
        * 1j
        * (R1m(psi, coefficients) + R1p(psi, coefficients))
        / np.sqrt(2)
    )


@lru_cache(maxsize=int(1e6))
def HSz(psi: UncoupledBasisState, coefficients: XConstants) -> UncoupledState:
    return -coefficients.D_TlF * R10(psi, coefficients)


# Old functions from Jakobs original Hamiltonian


def R10(psi: UncoupledBasisState, coefficients: XConstants) -> UncoupledState:
    amp1 = np.sqrt(2) * np.sqrt(
        (psi.J - psi.mJ) * (psi.J + psi.mJ) / (8 * psi.J**2 - 2)
    )
    ket1 = UncoupledBasisState(
        psi.J - 1,
        psi.mJ,
        psi.I1,
        psi.m1,
        psi.I2,
        psi.m2,
        Omega=psi.Omega,
        P=parity_X(psi.J - 1),
        electronic_state=psi.electronic_state,
    )
    amp2 = np.sqrt(2) * np.sqrt(
        (psi.J - psi.mJ + 1) * (psi.J + psi.mJ + 1) / (6 + 8 * psi.J * (psi.J + 2))
    )
    ket2 = UncoupledBasisState(
        psi.J + 1,
        psi.mJ,
        psi.I1,
        psi.m1,
        psi.I2,
        psi.m2,
        Omega=psi.Omega,
        P=parity_X(psi.J + 1),
        electronic_state=psi.electronic_state,
    )
    return UncoupledState([(amp1, ket1), (amp2, ket2)])


def R1m(psi: UncoupledBasisState, coefficients: XConstants) -> UncoupledState:
    amp1 = (
        -0.5
        * np.sqrt(2)
        * np.sqrt((psi.J + psi.mJ) * (psi.J + psi.mJ - 1) / (4 * psi.J**2 - 1))
    )
    ket1 = UncoupledBasisState(
        psi.J - 1,
        psi.mJ - 1,
        psi.I1,
        psi.m1,
        psi.I2,
        psi.m2,
        Omega=psi.Omega,
        P=parity_X(psi.J - 1),
        electronic_state=psi.electronic_state,
    )
    amp2 = (
        0.5
        * np.sqrt(2)
        * np.sqrt(
            (psi.J - psi.mJ + 1) * (psi.J - psi.mJ + 2) / (3 + 4 * psi.J * (psi.J + 2))
        )
    )
    ket2 = UncoupledBasisState(
        psi.J + 1,
        psi.mJ - 1,
        psi.I1,
        psi.m1,
        psi.I2,
        psi.m2,
        Omega=psi.Omega,
        P=parity_X(psi.J + 1),
        electronic_state=psi.electronic_state,
    )
    return UncoupledState([(amp1, ket1), (amp2, ket2)])


def R1p(psi: UncoupledBasisState, coefficients: XConstants) -> UncoupledState:
    amp1: float = (
        -0.5
        * np.sqrt(2)
        * np.sqrt((psi.J - psi.mJ) * (psi.J - psi.mJ - 1) / (4 * psi.J**2 - 1))
    )
    ket1 = UncoupledBasisState(
        psi.J - 1,
        psi.mJ + 1,
        psi.I1,
        psi.m1,
        psi.I2,
        psi.m2,
        Omega=psi.Omega,
        P=parity_X(psi.J - 1),
        electronic_state=psi.electronic_state,
    )
    amp2: float = (
        0.5
        * np.sqrt(2)
        * np.sqrt(
            (psi.J + psi.mJ + 1) * (psi.J + psi.mJ + 2) / (3 + 4 * psi.J * (psi.J + 2))
        )
    )
    ket2 = UncoupledBasisState(
        psi.J + 1,
        psi.mJ + 1,
        psi.I1,
        psi.m1,
        psi.I2,
        psi.m2,
        Omega=psi.Omega,
        P=parity_X(psi.J + 1),
        electronic_state=psi.electronic_state,
    )
    return UncoupledState([(amp1, ket1), (amp2, ket2)])


def HI1R(psi: UncoupledBasisState, coefficients: XConstants) -> UncoupledState:
    return com(I1z, R10, psi, coefficients) + (
        com(I1p, R1m, psi, coefficients) - com(I1m, R1p, psi, coefficients)
    ) / np.sqrt(2)


def HI2R(psi: UncoupledBasisState, coefficients: XConstants) -> UncoupledState:
    return com(I2z, R10, psi, coefficients) + (
        com(I2p, R1m, psi, coefficients) - com(I2m, R1p, psi, coefficients)
    ) / np.sqrt(2)


def Hc3_alt(psi: UncoupledBasisState, coefficients: XConstants) -> UncoupledState:
    return 5 * coefficients.c3 / coefficients.c4 * Hc4(
        psi, coefficients
    ) - 15 * coefficients.c3 / 2 * (
        com(HI1R, HI2R, psi, coefficients) + com(HI2R, HI1R, psi, coefficients)
    )


@lru_cache(maxsize=int(1e6))
def Hff_alt(psi: UncoupledBasisState, coefficients: XConstants) -> UncoupledState:
    return (
        Hrot(psi, coefficients)
        + Hc1(psi, coefficients)
        + Hc2(psi, coefficients)
        + Hc3_alt(psi, coefficients)
        + Hc4(psi, coefficients)
    )
